import json
from pathlib import Path

# hex to Lego RGB codes
# code messages are always shape (hexadecimal): 08004501xx00ffff
# (where xx is the tile code)
with open(Path(__file__).parent / Path("ALL_RGB_CODES.json")) as f:
    HEX_TO_RGB_TILE = {x[2]:x[1] for x in json.load(f)}

# hex to ground colors
# color messages are always shape (hexadecimal): 08004501ffffxx00
# (where xx is the color code)
HEX_TO_COLOR_TILE = {
    0x0c: "Purple",
    0x13: "White",
    0x15: "Red",
    0x17: "Blue",
    0x18: "Yellow",
    0x1a: "Black",
    0x25: "Green",
    0x38: "Nougat Brown",
    0x42: "Cyan",
    0x6a: "Brown"}

# hex to pants codes
HEX_TO_PANTS = {        # Pins
    0x00: "None",       #000000
    0x03: "Bee",        #000011
    0x05: "Luigi",      #000101
    0x06: "Frog",       #000110
    0x0a: "Tanooki",    #001010
    0x0c: "Propeller",  #001100
    0x11: "Cat",        #010001
    0x12: "Fire",       #010010
    0x14: "Penguin",    #010100
    0x18: "Peach",      #011000
    0x21: "Mario",      #100001
    0x22: "Builder"     #100010
}

HEX_TO_HUB_ACTIONS = {
    0x30: "Hub Will Switch Off",
    0x31: "Hub Will Disconnect"
}

HEX_TO_HUB_PROPERTIES = {
    0x01: "Advertising Name",
    0x02: "Button",
    0x03: "Firmware Ver.",
    0x04: "Hardware Ver.",
    0x05: "Signal Strength",
    0x06: "Battery Voltage",
    0x07: "Battery Type",
    0x08: "Manufacturer Name",
    0x09: "Radio Firmware Version",
    0x0A: "LEGO Wireless Protocol Version",
    0x0B: "System Type ID",
    0x0C: "H/W Network ID",
    0x0D: "Primary MAC Address",
    0x0E: "Secondary MAC Address",
    0x0F: "Hardware Network Family"
}

# BLE Connection
# https://github.com/bricklife/LEGO-Mario-Reveng
LEGO_CHARACTERISTIC_UUID = "00001624-1212-efde-1623-785feabcd123"
# Request Commmands
REQUEST_RGB_COMMAND = bytearray([
                                0x05,  # message length
                                0x00,  # unused
                                0x21,  # message type (21=Port Information Request)
                                0x01,  # port ID
                                0x00   # requested information type (0=Port Value)
                                ])
REQUEST_PANTS_COMMAND = bytearray([0x05, 0x00, 0x21, 0x02, 0x00])
REQUEST_IMU_COMMAND = bytearray([0x05, 0x00, 0x21, 0x00, 0x00])

VALID_PORT_MODES = {0:(0,1), 1:(0,1), 2:(0,), 3:(0,1,2,3), 4:(0,), 6:(0,)}

def pifs_command(port: int, mode: int, notifications: bool = True, delta_interval: int = 1):
    """Creates a PORT_INPUT_FORMAT_SETUP message according to the Lego
    Wireless Protocol: https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-input-format-setup-single

    Args:
    port (int): The designated Port.
                Port 0: Accelerometer
                Port 1: Camera
                Port 2: Binary (Pants)
                Port 3: ??
                Port 4: ??
    mode (int): The mode to set the port to. Available modes:
        Port 0: (0,1), Port 1: (0,1), Port 2: (0), Port 3: (0,1,2,3),
        Port 4: (0,1). Also see https://github.com/bricklife/LEGO-Mario-Reveng
    notifications (bool, optional): Whether to receive updates about every new
        value of the port. Defaults to True. If False, you'll need to manually
        request port values.
    delta_interval (int, optional): The necessary change in measured data to
        trigger a change in the sent value. A higher delta interval leads to
        less accurate data, but also less data traffic via bluetooth. I
        recommend 1 for every port except Port 0, which I recommend to set
        between 1 and 5 depending on your machine. Defaults to 1.

    Returns:
        bytearray: A bytearray to be sent to Lego Mario
    """
    if port not in (0,1,2,3,4):
        raise ValueError(f"Invalid Port, expected one of (0,1,2,3,4) got {port}")
    if mode not in VALID_PORT_MODES[port]:
        raise ValueError(
            f"Invalid mode {mode} for port {port}, allowed modes for this \
            port are: {VALID_PORT_MODES[port]}.")
    # Delta Interval
    try:
        int(delta_interval)
    except (TypeError, ValueError) as e:
        raise TypeError(f"Delta Interval must be castable to int, got \
                        {type(delta_interval)} instead") from e

    return bytearray([
                    0x0A,  # Message Length
                    0x00,  # Unused
                    0x41,  # Message Type
                    port,
                    mode,
                    int(delta_interval),
                    0x00,
                    0x00,
                    0x00,
                    int(bool(notifications))  # Notifications en/disabled
                    ])

# Subscribtion Commands
SUBSCRIBE_IMU_COMMAND = pifs_command(0, 0, delta_interval=2)
SUBSCRIBE_RGB_COMMAND = pifs_command(1, 0, delta_interval=1)
SUBSCRIBE_PANTS_COMMAND = pifs_command(2, 0)

MUTE_COMMAND = bytearray([
                        0x06, # message length
                        0x00, # unused, always 0
                        0x01, # message type (01 = Hub Properties)
                        0x12, # specify hub property (12 = volume)
                        0x01, # specify operation (1 = set new value)
                        0x00  # new value (0 = mute, 100 = full volume)
                        ])
DISCONNECT_COMMAND = bytearray([
                        0x04, # message length
                        0x00, # unused, always 0
                        0x02, # message type (02 = HUB Actions)
                        0x02, # specify action (02 = disconnect)
                        ])
TURN_OFF_COMMAND = bytearray([
                        0x04, # message length
                        0x00, # unused, always 0
                        0x02, # message type (02 = HUB Actions)
                        0x01, # specify action (01 = turn off)
                        ])

BINARY_GESTURES = { # most likely wrong
                0b0000000000000001: "Bump",
                0b0000000000000010: "Gesture2",
                0b0000000000000100: "Gesture4",
                0b0000000000001000: "Gesture8",
                0b0000000000010000: "Shake",
                0b0000000000100000: "Gesture32",
                0b0000000001000000: "Gesture64",
                0b0000000010000000: "Gesture128",
                0b0000000100000000: "Turning",
                0b0000001000000000: "Fastmove",
                0b0000010000000000: "Translation",
                0b0000100000000000: "HighFallCrash",
                0b0001000000000000: "DirectionChange",
                0b0010000000000000: "Reverse",
                0b0100000000000000: "Gesture16384",
                0b1000000000000000: "Jump"
}



"""
pygame_mario.py
This file implements class PygameMario that subclasses Mario to utilize pygame
events. It also implements AsyncClock, which behaves like a regular pygame
clock but also keeps Mario running.
Copyright (c) 2022 Jamin Kauf
"""
import os

os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"
import pygame
from pygame.locals import *
from pyLegoMario import Mario
import asyncio

ACC_EVENT = pygame.event.custom_type()
RGB_EVENT = pygame.event.custom_type()
PANTS_EVENT = pygame.event.custom_type()


def _acceleration_callback(mario: Mario, x: int, y: int, z: int):
    event = pygame.event.Event(ACC_EVENT, value=(x, y, z), sender=mario)
    pygame.event.post(event)


def _rgb_callback(mario: Mario, t: str):
    event = pygame.event.Event(RGB_EVENT, value=t, sender=mario)
    pygame.event.post(event)


def _pants_callback(mario: Mario, powerup: str):
    event = pygame.event.Event(PANTS_EVENT, value=powerup, sender=mario)
    pygame.event.post(event)


class AsyncClock:
    """Use this instead of pygame.time.Clock when using Lego Mario.
    Your AsyncClock object's tick method must be called once per frame,
    otherwise Mario cannot send events.
    """

    def __init__(self) -> None:
        self.clock = pygame.time.Clock()
        self.get_time = self.clock.get_time
        self.get_rawtime = self.clock.get_rawtime
        self.get_fps = self.clock.get_fps
        self._tick = self.clock.tick
        self.loop = asyncio.get_event_loop()
        self._tick_busy_loop = self.clock.tick_busy_loop

    def tick(self, framerate: int = 0) -> int:
        """Limits framerate by blocking, but uses spare time for async loop.

        Args:
            framerate (int, optional): Maximum desired framerate. If 0 or not
                provided, will not delay. Defaults to 0.

        Returns:
            int: The number of milliseconds since last call.
        """
        self.loop.run_until_complete(asyncio.sleep(0.01))
        return self._tick(framerate)

    def tick_busy_loop(self, framerate: int = 0) -> int:
        self.loop.run_until_complete(asyncio.sleep(0.01))
        return self._tick_busy_loop(framerate)


class PygameMario(Mario):
    """Connect to Lego Mario and enable it to post pygame events.
    Lego Mario's events will contain a sender (event.sender), which is the
    Lego Mario object, and a value (event.value), which will either be a
    string (in case of pants or camera data) or a tuple of integers (in
    case of acceleration data)."""

    def __init__(self, enable_acc_events: bool = True, enable_rgb_events: bool = True, enable_pants_events: bool = True, **kwargs) -> None:
        """
        Args:
            enable_acc_events (bool, optional): Whether to send acceleration
                events. event.value will be tuple(int, int, int).
                Be aware that this will send a lot of events.
                Defaults to True.
            enable_rgb_events (bool, optional): Whether to send rgb events
                like RGB tiles or ground color. event.value will be str.
                Defaults to True.
            enable_pants_events (bool, optional): Whether to send pants events.
                event.value will be str. Defaults to True.
        """
        kwargs.setdefault("do_log", False)
        kwargs.setdefault("default_volume", 0)
        super().__init__(**kwargs)
        ports_task = self._init_ports(enable_acc_events, enable_rgb_events, enable_pants_events)
        asyncio.get_event_loop().create_task(ports_task)

    async def _init_ports(self, acc_enabled: bool, rgb_enabled: bool, pants_enabled: bool) -> None:
        await self.await_connection()
        if acc_enabled:
            self.add_accelerometer_hooks(_acceleration_callback)
        else:
            await self.port_setup(0, 0, False)
        if rgb_enabled:
            self.add_tile_hooks(_rgb_callback)
        else:
            await self.port_setup(1, 0, False)
        if pants_enabled:
            self.add_pants_hooks(_pants_callback)
        else:
            await self.port_setup(2, 0, False)



from typing import Union
import tkinter as tk
from pathlib import Path
import asyncio
from PIL import ImageTk, Image
try:
    from .mario import Mario
    from .lego_mario_data import *
except ImportError:
    from mario import Mario
    from lego_mario_data import *

class MarioWindow(tk.Frame):
    """Object that creates a GUI for Lego Mario.
    """
    def __init__(self, mario_entity: Mario,
                 master: Union[tk.Widget, None] = None) -> None:
        """

        Args:
            mario_entity (Mario): Instance of Lego Mario to create the GUI for
            master (_type_, optional): parent tkinter widget. Uses current
                new Toplevel otherwise. Defaults to None.
        """
        self.mario = mario_entity
        if master:
            tk.Frame.__init__(self, master)
        else:
            tk.Frame.__init__(self, tk.Toplevel() if tk._default_root else None)
        self.master: tk.Toplevel
        # Window Setup
        self.master.minsize(644, 165)
        self.master.iconbitmap(Path(__file__).parent / "icon.ico")
        bg_image = Image.open(Path(__file__).parent / "background.png")
        self.bg_image_tk = ImageTk.PhotoImage(bg_image)
        background_label = tk.Label(self, image=self.bg_image_tk)
        background_label.place(x=0,y=0, relheight=1, relwidth=1)
        # Resizing Geometry
        self.pack(fill=tk.BOTH, expand=True)
        self.columnconfigure([0,1,2,3],weight=1)
        self.rowconfigure(1,weight=1)

        # Start Acceleration Data
        self.acceleration_frame = tk.Frame(self)
        self.acceleration_frame.grid(row=0, column=0)
        self.x_label = tk.Label(self.acceleration_frame, text="X")
        self.y_label = tk.Label(self.acceleration_frame, text="Y")
        self.z_label = tk.Label(self.acceleration_frame, text="Z")
        self.x_label.grid(row=0, column=0)
        self.y_label.grid(row=0, column=1)
        self.z_label.grid(row=0, column=2)
        # StringVars Receiving Data
        self.x_acceleration_text = tk.StringVar()
        self.y_acceleration_text = tk.StringVar()
        self.z_acceleration_text = tk.StringVar()
        # Create Entries Using StringVars
        self.x_acceleration_box = tk.Entry(self.acceleration_frame, width=4,
                    state="readonly", textvariable=self.x_acceleration_text)
        self.y_acceleration_box = tk.Entry(self.acceleration_frame, width=4,
                    state="readonly", textvariable=self.y_acceleration_text)
        self.z_acceleration_box = tk.Entry(self.acceleration_frame, width=4,
                    state="readonly", textvariable=self.z_acceleration_text)
        self.x_acceleration_box.grid(row=1, column=0)
        self.y_acceleration_box.grid(row=1, column=1)
        self.z_acceleration_box.grid(row=1, column=2)
        # Hook Fields to Mario
        self.mario.add_accelerometer_hooks(self._input_acceleration_data)
        # End Acceleration Data

        # Start Pants Data
        # Static container (invisible frame) and description label
        self.pants_frame = tk.Frame(self)
        self.pants_label = tk.Label(self.pants_frame, text="Pants")
        # Variable and Entry for Displaying Variable
        self.pants_text_var = tk.StringVar()
        box_width = max([len(powerup) for powerup in HEX_TO_PANTS.values()])
        self.pants_box = tk.Entry(self.pants_frame, width=box_width,
                                  state="readonly",
                                  textvariable=self.pants_text_var)
        
        self.pants_frame.grid(row=0, column=2)
        self.pants_label.grid(row=0)
        self.pants_box.grid(row=1)
        # Hook Field to Mario
        self.mario.add_pants_hooks(self._input_pants_data)
        # End Pants Data

        # Start RGB Data
        self.rgb_frame = tk.Frame(self)
        self.rgb_label = tk.Label(self.rgb_frame, text="RGB/Tile")
        self.rgb_text = tk.StringVar()
        possible_values = (HEX_TO_RGB_TILE | HEX_TO_COLOR_TILE).values()
        box_width = max([len(value) for value in possible_values])
        self.rgb_box = tk.Entry(self.rgb_frame, width=box_width,
                               state='readonly', textvariable=self.rgb_text)
        self.rgb_frame.grid(row=0, column=1)
        self.rgb_label.grid(row=0)
        self.rgb_box.grid(row=1)
        # Hook Field to Mario
        self.mario.add_tile_hooks(self._input_rgb_data)
        # End RGB Data

        # Scale for Adjusting Volume
        self.volumeFrame = tk.Frame(self, bg="#5c94fc")
        self.volumeScale = tk.Scale(self.volumeFrame,
                                    from_=0, to=100, orient=tk.HORIZONTAL,
                                    command=self._set_mario_volume,
                                    label="Volume", highlightthickness=0)
        if not self.mario.default_volume is None:
            self.volumeScale.set(self.mario.default_volume)
        else:
            self.volumeScale.set(100)
        self.volumeFrame.grid(row=0, column=3)
        self.volumeScale.grid(row=0, column=0)

        # Logging Data
        self.logText = tk.StringVar()
        self.logBox = tk.Text(self, state=tk.DISABLED, width=80)
        self.logBox.grid(row=1, columnspan=6, sticky=tk.NSEW)
        self.mario.add_log_hooks(self._input_log_data)

        # Start Buttons
        # Button Container
        self.buttonFrame = tk.Frame(self, bg="#5c94fc")
        self.buttonFrame.grid(row=2, column=0, columnspan=2, sticky=tk.EW)
        self.buttonFrame.columnconfigure([0,1,2],weight=1)

        # This Button will also handle disconnecting
        self.connectButton = tk.Button(self.buttonFrame, text="Connect",
                                       command=self._dis_connect_mario)
        self.connectButton.grid(row=0, column=0)

        self.quit_button = tk.Button(self.buttonFrame, text='Quit',
                                     command=self.quit)
        self.quit_button.grid(row=0, column=1)

        self.turnOffButton = tk.Button(self.buttonFrame, text="Turn Off",
                                       command=self._turn_mario_off)
        self.turnOffButton.grid(row=0, column=2)
        # End Buttons

        # Checkbox for AutoReconnect
        self.reconnectVar = tk.IntVar(self, value=1) 
        self.reconnectCheckBox = tk.Checkbutton(self, text="Auto-Reconnect",
            variable=self.reconnectVar, command=self._set_auto_reconnect)
        self.reconnectCheckBox.grid(row=2, column=2)

        # Frame for port formatting
        self.portFormatFrame = tk.Frame(self, bg="#5c94fc")
        self.portFormatFrame.grid(row=3, column=3)

        # Checkbox for notifications
        self.notificationVar = tk.IntVar(self, value=1)
        self.notificationCheckBox = tk.Checkbutton(
            self.portFormatFrame, variable=self.notificationVar,
            text="Notifications")
        self.notificationCheckBox.grid(column=1, row=0)

        # OptionMenu for port modes
        self.port_mode_variable = tk.IntVar(self)
        options = ("0","1")
        self.mode_menu = tk.OptionMenu(self.portFormatFrame,
                                        self.port_mode_variable,
                                        *options)
        self.mode_menu.config(highlightthickness=0)
        self.port_mode_variable.set(0)
        self.mode_menu.grid(column=0, row=0)

        # Button for configuring port format
        self.portFormatButton = tk.Button(self.portFormatFrame, 
                                        text="Update Port", 
                                        command=self._set_port_input_format)
        self.portFormatButton.grid(column=2, row=0)

        # Request Port Values
        self.portVar = tk.IntVar(self, value=1)
        # Frame for All Content Regarding Port Values
        self.portFrame = tk.Frame(self)
        # Button to Request Value
        self.request_port_button = tk.Button(self.portFrame, 
            text="Request Value", command=self._request_port)
        # Radiobutton for Each Port (Grid All to portFrame)
        # no Port 0/5 (accelerometer isn't printed to Log, Port 5 doesn't exist)
        for i, port in enumerate((1,2,3,4,6)):
            radiobutton = tk.Radiobutton(
                self.portFrame, variable=self.portVar, value=port,
                text=f"Port {port}", command=self._update_mode_menu)
            radiobutton.grid(row=i // 3, column=i % 3)
        self.portFrame.grid(row=2, column=3)
        self.request_port_button.grid(row=1, column=2)

        # display and update the window
        asyncio.get_event_loop().create_task(self._run_window())

    def _update_mode_menu(self) -> None:
        """Update the port mode menu with the available port modes for the
        selected port. Called every time a different port gets selected.
        """
        new_choices = VALID_PORT_MODES[self.portVar.get()]

        # change menu accordingly
        self.mode_menu['menu'].delete(0, tk.END)
        for choice in new_choices:
            self.mode_menu['menu'].add_command(
                label=str(choice), 
                command=tk._setit(self.port_mode_variable, str(choice)))

        # only reset selection if previously selected mode isn't valid anymore
        if not int(self.port_mode_variable.get()) in new_choices:
            self.port_mode_variable.set(0)

    def _set_port_input_format(self) -> None:
        task = self.mario.port_setup(
            port=self.portVar.get(),
            mode=self.port_mode_variable.get(),
            notifications=bool(self.notificationVar.get()))
        asyncio.create_task(task)


    def _dis_connect_mario(self) -> None:
        """Connects Mario if Mario isn't running. Disconnect Mario if Mario is
        connected. Passes otherwise (e.g. running but not connected yet)
        """
        if not self.mario.run:
            asyncio.create_task(self.mario.connect())
        elif self.mario.client:
            asyncio.create_task(self.mario.disconnect())

    def _set_mario_volume(self, new_volume: str):
        """Creates asyncio task of Mario's set_volume Coroutine.
        The new volume is the current value of self.volumeVar
        (set by the volumeScale).

        Args:
            new_volume (str): the new volume as provided by tk.Scale
        """
        self.mario.default_volume = int(new_volume)
        self.mario.set_volume(self.mario.default_volume)
    
    def _set_auto_reconnect(self):
        """Sets Mario's ._autoReconnect attribute to the current value of 
        self.reconnectVar (determined by self.reconnectCheckBox)."""
        self.mario.auto_reconnect = bool(self.reconnectVar.get())

    def _turn_mario_off(self):
        """Creates asyncio task of Mario's .turn_off() Coroutine.
        """
        asyncio.create_task(self.mario.turn_off())

    def _request_port(self) -> None:
        """Requests the currently selected port of Mario.
        """
        port_id = self.portVar.get()
        loop = asyncio.get_event_loop()
        loop.create_task(self.mario.request_port_value(port_id))

    def _input_log_data(self, sender: Mario, msg: str) -> None:
        """Function to display log messages in GUI

        Args:
            sender (Mario): The Mario entity that sends the data.
                Used for compatibility with Mario's event hooks.
            msg (str): Log Message
        """
        assert sender == self.mario
        try:
            content, hex_msg = msg.split(", Hex: ")
            padded_width = self.logBox['width'] - len(hex_msg)
            msg = f"{content.ljust(padded_width)}{hex_msg}"
        except ValueError:
            # unable to split/format: leave message untouched
            pass
        # Insert Message Into Log Textbox (excluding acceleration data)
        if not msg.startswith("X: "):
            self.logBox['state'] = tk.NORMAL
            self.logBox.insert(tk.END, f"\n{msg}")
            self.logBox['state'] = tk.DISABLED
            self.logBox.see(tk.END)  # scroll down

    def _input_acceleration_data(self, sender: Mario,
                                 x: int, y: int, z: int) -> None:
        """Hook for acceleration data to be displayed on GUI

        Args:
            sender (Mario): The Mario entity that sends the data.
                Only used for compatibility with mario's event hooks.
            x (int): acceleration data in x direction
            y (int): acceleration data in y direction
            z (int): acceleration data in z direction
        """
        assert sender == self.mario
        self.x_acceleration_text.set(str(x))
        self.y_acceleration_text.set(str(y))
        self.z_acceleration_text.set(str(z))

    def _input_pants_data(self, sender: Mario, pants: str) -> None:
        """Hook for pants data to be displayed on GUI

        Args:
            sender (Mario): The Mario entity that sends the data.
                Only used for compatibility with mario's event hooks.
            pants (str): The type of pants mario is wearing. See
                LEGO_MARIO_DATA.py for more info
        """
        assert sender == self.mario
        self.pants_text_var.set(pants)

    def _input_rgb_data(self, sender: Mario, color_or_tile: str) -> None:
        """Hook for rgb/tile data to be displayed on GUI

        Args:
            sender (Mario): The Mario entity that sends the data.
                Only used for compatibility with mario's event hooks.
            color_or_tile (str): String of the color or tile.
                See LEGO_MARIO_DATA.py for more info
        """
        assert sender == self.mario
        self.rgb_text.set(color_or_tile)

    def quit(self) -> None:
        """Destroys the window and removes Mario's event hooks. 
        Mario remains connected. Call Mario.disconnect() to disconnect.
        """
        # Remove Hooks from Mario
        self.mario.remove_hooks((
            self._input_acceleration_data, self._input_pants_data, 
            self._input_rgb_data, self._input_log_data))
        self.mario.auto_reconnect = False
        # Close Window
        try:
            self.master.destroy()
        except tk.TclError as e:
            pass

    async def _run_window(self, interval: float = 0.05) -> None:
        """Endless loop that keeps the window running, updating it every
        INTERVAL seconds.\n
        This loop is also responsible for enabling and disabling GUI functions
        depending on Mario's connection status

        Args:
            interval (float, optional): Time interval between window updates.
            Necessary to handle Mario. Defaults to 0.05.
        """
        try:
            while True:
                # Mario is connected and running
                if self.mario.is_connected:
                    self.request_port_button.config(state=tk.NORMAL)
                    self.portFormatButton.config(state=tk.NORMAL)
                    self.turnOffButton.config(state=tk.NORMAL)
                    self.connectButton.config(text="Disconnect", state=tk.NORMAL)
                    self.master.title(f"Lego Mario - {self.mario.client.address}")
                # Mario is disconnected and not trying to connect
                elif not self.mario.run:
                    self.request_port_button.config(state=tk.DISABLED)
                    self.portFormatButton.config(state=tk.DISABLED)
                    self.turnOffButton.config(state=tk.DISABLED)
                    self.master.title("Lego Mario - Not Connected")
                    self.connectButton.config(text="Connect", state=tk.NORMAL)
                # Mario is running, but not connected (trying to connect)
                else:
                    self.request_port_button.config(state=tk.DISABLED)
                    self.portFormatButton.config(state=tk.DISABLED)
                    self.turnOffButton.config(state=tk.DISABLED)
                    self.master.title("Lego Mario - Connecting...")
                    self.connectButton.config(state=tk.DISABLED)
                # tkinter's update function
                self.update()
                await asyncio.sleep(interval)
        except tk.TclError as e:
            self.quit()  # remove event hooks in case of crash
            if "application has been destroyed" not in e.args[0] and "invalid command name" not in e.args[0]:
                print(e.args)  # debug


"""
MARIO.PY
###################################################################################
MIT License
Copyright (c) 2022 Bruno Hautzenberger, Jamin Kauf
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import asyncio
from typing import Any, Callable, Iterable, Union
from bleak import BleakScanner, BleakClient, BleakError

try:
    from .lego_mario_data import (
        HEX_TO_RGB_TILE,
        HEX_TO_COLOR_TILE,
        HEX_TO_PANTS,
        HEX_TO_HUB_ACTIONS,
        HEX_TO_HUB_PROPERTIES,
        BINARY_GESTURES,
        LEGO_CHARACTERISTIC_UUID,
        SUBSCRIBE_IMU_COMMAND,
        SUBSCRIBE_PANTS_COMMAND,
        SUBSCRIBE_RGB_COMMAND,
        DISCONNECT_COMMAND,
        pifs_command,
        TURN_OFF_COMMAND,
        MUTE_COMMAND,
        REQUEST_RGB_COMMAND,
    )
except ImportError:
    from lego_mario_data import (
        HEX_TO_RGB_TILE,
        HEX_TO_COLOR_TILE,
        HEX_TO_PANTS,
        HEX_TO_HUB_ACTIONS,
        HEX_TO_HUB_PROPERTIES,
        BINARY_GESTURES,
        LEGO_CHARACTERISTIC_UUID,
        SUBSCRIBE_IMU_COMMAND,
        SUBSCRIBE_PANTS_COMMAND,
        SUBSCRIBE_RGB_COMMAND,
        DISCONNECT_COMMAND,
        pifs_command,
        TURN_OFF_COMMAND,
        MUTE_COMMAND,
        REQUEST_RGB_COMMAND,
    )


class Mario:
    """Object to control and monitor a Lego Mario via Bluetooth.

    Attributes
    ----------
    do_log: bool
        If True, log messages will be printed to stdout.
    pants: str | None
        Value of most recent pants value
    ground: str | None
        Value of most recent camera/rgb value
    acceleration: tuple[int, int, int] | None
        Value of most recent acceleration value
    auto_reconnect: bool
        Whether .connect() should be called after disconnecting
    run: bool
        TBD
    client: BleakClient | None
        The bluetooth client that communicates with Mario
    default_volume: int | None
        The % volume that Mario will be set to after (re)connecting.
    _accelerometer_hooks: list[(Mario, int, int, int) -> None]
        List of callback functions for accelerometer updates.
    _pants_event_hooks: list[(Mario, str) -> None]
        List of callback functions for pants updates.
    _tile_event_hooks: list[(Mario, str) -> None]
        List of callback functions for camera/rgb updates.
    _log_event_hooks: list[(Mario, str) -> None]
        List of callback functions for log messages.
    _all_hooks: tuple[list[callbacks]]
        tuple that contains all of the previous lists of callback functions

    Methods
    -------
    add_accelerometer_hooks: (Callable | list[Callable]) -> None
        Adds the given function(s) as callback functions for accelerometer data
    add_pants_hooks: (Callable | list[Callable]) -> None
        Adds the given function(s) as callback functions for pants data
    add_tile_hooks: (Callable | list[Callable]) -> None
        Adds the given function(s) as callback functions for tile data
    add_log_hooks: (Callable | list[Callable]) -> None
        Adds the given function(s) as callback functions for log calls
    remove_hooks: (list[Any] | Callable) -> None
        Removes the given object(s) from all hook lists.
    log: (str) -> None
        logs the message to stdout if self.do_log is true. Also passes message
        to all callback functions in self._log_hooks.
    connect: () -> Coroutine
        Searches for Lego Mario objects via bluetooth and tries to connect.
        Needs to be awaited.
    set_volume: (int) -> None
        If mario is connected, sets volume to the % volume given. Also sets
        self.default_volume to keep volume persistent with reconnects.
    """

    def __init__(
        self,
        do_log: bool = True,
        accelerometer_hooks: Union[
            Callable[["Mario", int, int, int], Any],
            Iterable[Callable[["Mario", int, int, int], Any]],
        ] = [],
        tile_event_hooks: Union[
            Callable[["Mario", str], Any], Iterable[Callable[["Mario", str], Any]]
        ] = [],
        pants_event_hooks: Union[
            Callable[["Mario", str], Any], Iterable[Callable[["Mario", str], Any]]
        ] = [],
        log_event_hooks: Union[
            Callable[["Mario", str], Any], Iterable[Callable[["Mario", str], Any]]
        ] = [],
        default_volume: Union[int, None] = None,
    ) -> None:
        """
        Args:
            do_log (bool, optional): Enables Logs to Stdout. Defaults to True.

            accelerometerEventHooks (func or list of functions, optional):
                Event Hook(s) that should be called every time new
                accelerometer data is received.
                Functions need to take four inputs
                (sender: Mario, x: int, y: int, z: int). Defaults to None.

            tileEventHooks (func or list of functions, optional):
                Event Hook(s) that should be called every time new camera
                data is received. Functions need to take two inputs:
                (sender: Mario, ground: str). Defaults to None.

            pantsEventHooks (func or list of functions, optional):
                Event hook(s) that will be called every time new pants data
                is received. Functions need to take two inputs:
                (sender: Mario, pants: str). Defaults to None.

            logEventHooks (func or list of functions, optional): Event Hook(s)
                that will be called every time something gets logged. Functions
                need to take two inputs: (sender: Mario, msg: str).
                Defaults to None.

            defaultVolume (func or list of functions, optional): Volume (0-100)
                that will be set every time Mario reconnects. If not provided,
                will not adjust volume. Defaults to None.
        """

        self.do_log = do_log  # output logs to stdout if True
        self.run = False
        self.auto_reconnect = True  # handles reconnection on disconnect
        self.client: BleakClient | None = None
        self.default_volume = default_volume  # if None, volume won't be changed

        # values to keep most recent event in memory
        self.pants: str | None = None
        self.ground: str | None = None
        self.acceleration: tuple[int, int, int] | None = None
        self.recent_tile: str | None = None

        self._accelerometer_hooks: list[Callable[[Mario, int, int, int], Any]] = []
        self._tile_event_hooks: list[Callable[[Mario, str], Any]] = []
        self._pants_event_hooks: list[Callable[[Mario, str], Any]] = []
        self._log_event_hooks: list[Callable[[Mario, str], Any]] = []
        self._all_hooks = (
            self._accelerometer_hooks,
            self._pants_event_hooks,
            self._tile_event_hooks,
            self._log_event_hooks,
        )

        self.add_accelerometer_hooks(accelerometer_hooks)
        self.add_tile_hooks(tile_event_hooks)
        self.add_pants_hooks(pants_event_hooks)
        self.add_log_hooks(log_event_hooks)

        try:  # if event loop exists, use that one
            asyncio.get_event_loop().create_task(self.connect())
        except RuntimeError:  # otherwise, create a new one
            asyncio.set_event_loop(asyncio.SelectorEventLoop())
            asyncio.get_event_loop().create_task(self.connect())

    def log(self, msg: str, end: str = "\n") -> None:
        """Log any message to stdout and call all assigned LogEvent handlers.

        Args:
            msg (object): Any printable object.
            end (str, optional): Same as end in print(). Defaults to "\n".
        """
        for func in self._log_event_hooks:
            func(self, msg)
        if self.do_log:
            address = "Not Connected" if not self.client else self.client.address
            print((f"\r{address}: {msg}").ljust(100), end=end)

    def add_log_hooks(
        self,
        funcs: Union[
            Callable[["Mario", str], Any], Iterable[Callable[["Mario", str], Any]]
        ],
    ) -> None:
        """Adds function(s) as event hooks for updated tile or color values.

        Args:
            funcs (func or list of functions): callback functions must take
                (Mario, str) as input.
        """
        if callable(funcs):
            self._log_event_hooks.append(funcs)
        elif hasattr(funcs, "__iter__"):
            for hook_function in funcs:
                self.add_log_hooks(hook_function)

    def add_tile_hooks(
        self,
        funcs: Union[
            Callable[["Mario", str], Any], Iterable[Callable[["Mario", str], Any]]
        ],
    ) -> None:
        """Adds function(s) as event hooks for updated tile or color values.

        Args:
            funcs (func or list of functions): callback functions must take
                (Mario, str) as input.
        """
        if callable(funcs):
            self._tile_event_hooks.append(funcs)
        elif hasattr(funcs, "__iter__"):
            for hook_function in funcs:
                self.add_tile_hooks(hook_function)

    def add_accelerometer_hooks(
        self,
        funcs: Union[
            Callable[["Mario", int, int, int], Any],
            Iterable[Callable[["Mario", int, int, int], Any]],
        ],
    ) -> None:
        """Adds function(s) as event hooks for updated accelerometer values.

        Args:
            funcs (function or list of functions): callback function(s) take
            input as (Mario, int, int, int).
        """
        if callable(funcs):
            self._accelerometer_hooks.append(funcs)
        elif hasattr(funcs, "__iter__"):
            for hook_function in funcs:
                self.add_accelerometer_hooks(hook_function)

    def add_pants_hooks(
        self,
        funcs: Union[
            Callable[["Mario", str], Any], Iterable[Callable[["Mario", str], Any]]
        ],
    ) -> None:
        """Adds function(s) as event hooks for updated pants values.

        Args:
            funcs (func or list of functions): callback function(s) take
                input as (Mario, str).
        """
        if callable(funcs):
            self._pants_event_hooks.append(funcs)
        elif hasattr(funcs, "__iter__"):
            for hook_function in funcs:
                self.add_pants_hooks(hook_function)

    def remove_hooks(
        self, funcs: Union[Callable[..., Any], Iterable[Callable[..., Any]]]
    ) -> None:
        """Removes function(s) as event hooks.
            Note that this is without consideration for the type of hook.

        Args:
            funcs (Union[ Callable[[Any], Any], Iterable[Callable[[Any], Any]]]):
                callable or iterable of callable.
        """
        if callable(funcs):
            for hooktype in self._all_hooks:
                if funcs in hooktype:
                    hooktype.remove(funcs)
        elif hasattr(funcs, "__iter__"):
            for hook_function in funcs:
                self.remove_hooks(hook_function)

    def _call_tile_hooks(self, tile: str) -> None:
        self.ground = tile
        for func in self._tile_event_hooks:
            func(self, tile)

    def _call_accelerometer_hooks(self, x: int, y: int, z: int) -> None:
        self.acceleration = (x, y, z)
        for func in self._accelerometer_hooks:
            func(self, x, y, z)

    def _call_pants_hooks(self, powerup: str) -> None:
        self.pants = powerup
        for func in self._pants_event_hooks:
            func(self, powerup)

    def _handle_events(self, sender: int, data: bytearray) -> None:
        """Handles bluetooth notifications.

        Decodes the sent data and calls Mario's appropriate event hooks.

        Args:
            sender (int): Only necessary for bleak compatibility
            data (bytearray): The data of the notification
        """
        hex_data = data.hex()
        # Port Value
        if data[2] == 0x45:
            # Camera Sensor Data
            if data[3] == 0x01:
                if data[4] == data[5] == data[6] == data[7] == 0xFF:
                    self.log(f"IDLE?, Hex: {hex_data}")
                elif data[4] == data[5] == 0xFF:
                    # Ground Colors
                    color = HEX_TO_COLOR_TILE.get(
                        data[6], f"Unkown Color: {hex(data[6])}"
                    )
                    self.log(f"{color} Ground, Hex: {hex_data}")
                    self._call_tile_hooks(color)
                else:
                    # RGB code
                    tile_code = int.from_bytes(data[4:], "big")
                    tile_name = HEX_TO_RGB_TILE.get(
                        tile_code, f"Unkown Tile Code: {hex(tile_code)}"
                    )
                    self.recent_tile = tile_name
                    self.log(f"{tile_name} Tile, Hex: {hex_data}")
                    self._call_tile_hooks(tile_name)

            # Accelerometer data
            elif data[3] == 0x00:
                # Gesture Mode - experimental, likely not accurate
                if data[4:6] == data[6:]:
                    gesture = ""
                    integer_data = int.from_bytes(data[4:6], "big")
                    for binary, name in BINARY_GESTURES.items():
                        if integer_data & binary:
                            gesture += name
                    self.log(gesture)

                # RAW Mode
                else:
                    try:
                        x = int(signed(data[4]))
                        y = int(signed(data[5]))
                        z = int(signed(data[6]))
                    except IndexError:
                        error_msg = (
                            f"Message length is {len(data)}, expected 6."
                            "The most likely cause is outdated software. To update, "
                            "connect your Lego Mario to the Lego Mario smartphone app."
                        )
                        self.log(error_msg)
                        raise ValueError(error_msg)
                    self.log(f"X: {x} Y: {y} Z: {z}", end="")
                    self._call_accelerometer_hooks(x, y, z)

            # Pants data
            elif data[3] == 0x02:
                pants = HEX_TO_PANTS.get(data[4], "Unkown")
                binary_pants = bin(data[4])
                self.log(
                    f"{pants} Pants, "
                    f"Pants-Only Binary: {binary_pants},"
                    f"Hex: {hex_data}"
                )
                self._call_pants_hooks(pants)
            # Port 3 data - uncertain about all of it
            elif data[3] == 0x03:
                if data[4] == 0x13 and data[5] == 0x01:
                    tile_name = HEX_TO_RGB_TILE.get(data[6], "Unkown Tile")
                    self.log(f"Port 3: Jumped on {tile_name}, Hex: {hex_data}")
                else:
                    # TBD
                    self.log(
                        f"Unknown value from port 3: {data[4:].hex()}, "
                        f"Hex: {hex_data}"
                    )
            else:
                self.log(
                    f"Unknown value from port {data[3]}: "
                    f"{data[4:].hex()}, Hex: {hex_data}"
                )

        # other technical messages
        elif data[2] == 0x02:  # Hub Actions
            action = HEX_TO_HUB_ACTIONS.get(
                data[3], f"Unkown Hub Action, Hex: {hex_data}"
            )
            self.log(f"{action}, Hex: {hex_data}")
            if data[3] == 0x31:  # 0x31 = Hub Will Disconnect
                asyncio.get_event_loop().create_task(self.disconnect())
        elif data[2] == 0x04:  # Hub Attached I/O
            if data[4]:
                self.log(f"Port {data[3]} got attached, Hex: {hex_data}")
            else:
                self.log(
                    f"Port {data[3]} got detached, "
                    f"this shouldn't happen. Hex: {hex_data}"
                )
        elif data[2] == 0x47:  # Port Input Format Handshake
            self.log(
                f"Port {data[3]} changed to mode {data[4]} "
                f"with{'out' if not data[9] else ''} notifications, "
                f"Hex: {hex_data}"
            )
        elif data[2] == 0x01 and data[4] == 0x06:
            hub_property = HEX_TO_HUB_PROPERTIES.get(data[3], "Unknown Property")
            self.log(
                f"Hub Update About {hub_property}: "
                f"{data[5:].hex()}, "
                f"Hex: {hex_data}"
            )
        else:  # Other
            self.log(
                f"Unknown message - check Lego Wireless Protocol, " f"Hex: {hex_data}"
            )

    async def connect(self) -> bool:
        self.run = True
        retries = 0
        while self.run:
            retries += 1
            if retries > 3:
                self.log("Stopped after 3 attempts, disconnecting...")
                break
            self.log("Searching for device...")
            try:
                devices = await BleakScanner.discover()
            except OSError as e:
                raise OSError(
                    "Can't use device - make sure your device"
                    f" supports Bluetooth and turn it on.\n{e}"
                )
            for d in devices:
                if d.name and (
                    d.name.lower().startswith("lego luigi")
                    or d.name.lower().startswith("lego mario")
                    or d.name.lower().startswith("lego peach")
                ):
                    try:
                        client = BleakClient(d)
                        await client.connect()
                        self.client = client
                        self.log(f"Mario Connected: {client.address}")

                        # subscribe to events
                        await client.start_notify(
                            LEGO_CHARACTERISTIC_UUID, self._handle_events
                        )
                        await asyncio.sleep(0.1)
                        await client.write_gatt_char(
                            LEGO_CHARACTERISTIC_UUID, SUBSCRIBE_IMU_COMMAND
                        )
                        await asyncio.sleep(0.1)
                        await client.write_gatt_char(
                            LEGO_CHARACTERISTIC_UUID, SUBSCRIBE_RGB_COMMAND
                        )
                        await asyncio.sleep(0.1)
                        await client.write_gatt_char(
                            LEGO_CHARACTERISTIC_UUID, SUBSCRIBE_PANTS_COMMAND
                        )

                        asyncio.get_event_loop().create_task(
                            self._check_connection_loop()
                        )

                        if not self.default_volume is None:
                            self.set_volume(self.default_volume)
                        return True
                    except Exception as ex:
                        if "-2140864507" in ex:
                            self.log(
                                "Mario's newest Firmware needs to be paired with your computer once before connecting. If you're on mac, please do so manually in settings. On Windows, this will now be done automatically. Try reconnecting"
                            )
                            await client.pair()
                        self.log(f"Error connecting: {ex}")
                        await self.disconnect()
                        return False
        await self.disconnect()
        return False

    async def request_port_value(self, port: int = 0) -> None:
        """Method for sending request for color sensor port value to Mario.
        Default port is 0.
        0 - Accelerometer
        1 - Camera
        2 - Pants
        3 - unknown
        4 - unknown
        6 - voltage?
        Response will be sent to event handlers.

        Args:
            port (int, optional): Port to request value from. Defaults to 0.
        """
        assert port in (0, 1, 2, 3, 4, 6), "Use a supported port (0,1,2,3,4,6)"
        if self.client:
            try:
                command = REQUEST_RGB_COMMAND
                command[3] = port
                command = bytearray(command)
                await self.client.write_gatt_char(LEGO_CHARACTERISTIC_UUID, command)
            except (OSError, BleakError):
                self.log("Connection error while requesting port value")
                await self.disconnect()

    def set_volume(self, new_volume: int) -> None:
        """Sets mario's volume to the specified volume.

        Args:
            new_volume (int): Percentage of maximum volume.
                Values <0 or >100 will be set to 0 or 100 respectively."""
        new_volume = min(max(new_volume, 0), 100)
        if self.client:
            try:
                command = bytearray([*MUTE_COMMAND[:5], new_volume])
                asyncio.get_event_loop().create_task(
                    self.client.write_gatt_char(LEGO_CHARACTERISTIC_UUID, command)
                )
            except (OSError, BleakError):
                self.log("Connection error while setting volume")
                asyncio.get_event_loop().create_task(self.disconnect())

    async def port_setup(
        self, port: int, mode: int, notifications: bool = True
    ) -> None:
        """Configures the settings of one of Mario's ports.
        Sends a message to Mario that configures the way one of its ports
        communicates.

        Args:
        port (int): The designated Port.
            Port 0: Accelerometer
            Port 1: Camera
            Port 2: Binary (Pants)
            Port 3: ??
            Port 4: ??
        mode (int): The mode to set the port to.
            Available modes:
                Port 0: (0,1),
                Port 1: (0,1),
                Port 2: (0),
                Port 3: (0,1,2,3),
                Port 4: (0,1).
            Also see https://github.com/bricklife/LEGO-Mario-Reveng
        notifications (bool, optional): Whether to receive updates about
            new values of the port. Defaults to True.
            If False, you'll need to manually request port values.
        """
        await self.await_connection()
        if self.client:
            try:
                command = pifs_command(port, mode, notifications)
                await self.client.write_gatt_char(LEGO_CHARACTERISTIC_UUID, command)
            except (OSError, BleakError):
                self.log("Connection error while setting up port")
                await self.disconnect()

    async def _check_connection_loop(self) -> None:
        while self.client:
            try:
                if not self.client.is_connected:
                    self.log("Disconnect detected during connection check")
                    await self.disconnect()
                await asyncio.sleep(3)
            except (OSError, BleakError):
                self.log("Error during connection check")
                await self.disconnect()

    async def disconnect(self) -> None:
        try:
            self.log("Disconnecting... ")
            if self.client:
                await self.client.write_gatt_char(
                    LEGO_CHARACTERISTIC_UUID, DISCONNECT_COMMAND
                )
                await self.client.disconnect()
                self.client = None
        except (OSError, BleakError):
            self.log("Connection error while disconnecting")
            self.client = None
        if self.auto_reconnect:
            asyncio.get_event_loop().create_task(self.connect())
        else:
            self.run = False

    async def turn_off(self) -> None:
        if not self.client:
            return
        try:
            self.log("Turning Off... ")
            await self.client.write_gatt_char(
                LEGO_CHARACTERISTIC_UUID, TURN_OFF_COMMAND
            )
            await self.disconnect()
        except (OSError, BleakError):
            self.log("Connection error while turning off")
            await self.disconnect()

    async def await_connection(self):
        while not self.is_connected:
            await asyncio.sleep(0.5)
        return

    @property
    def is_connected(self):
        if self.client:
            return self.client.is_connected
        return False

    def __str__(self) -> str:
        if self.is_connected:
            assert self.client
            return f"Mario at {self.client.address}"
        else:
            return "Mario - not connected"


def signed(char):
    return char - 256 if char > 127 else char


def run():
    """Runs the asyncio event loop until until all tasks are done."""
    while asyncio.all_tasks(loop=asyncio.get_event_loop()):
        asyncio.get_event_loop().run_until_complete(
            asyncio.gather(*asyncio.all_tasks(loop=asyncio.get_event_loop()))
        )



from setuptools import setup, find_packages

VERSION = '1.1'

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='pyLegoMario',
    version=VERSION,
    author='Jamin Kauf, Bruno Hautzenberger',
    author_email='jamin.kauf@yahoo.de',
    description='Module for handling connections with the Lego Mario toy',
    long_description_content_type='text/markdown',
    long_description=long_description,
    packages=find_packages(),
    install_requires=['bleak', 'pathlib', 'asyncio', 'pillow', 'pygame'],
    keywords=['lego', 'python', 'super mario', 'lego mario', 'bluetooth'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Operating System :: Microsoft :: Windows',
        'License :: OSI Approved :: MIT License'
    ],
    python_requires='>=3.9',
    project_urls = {'repository': r'https://github.com/Jackomatrus/pyLegoMario'},
    include_package_data=True
)